import logging

from pyqalx.core.adapters.adapter import QalxSignalAdapter, \
    QalxUpdateStatusAdapter
from pyqalx.core.entities.worker import Worker
from pyqalx.core.signals import QalxWorkerSignal

logger = logging.getLogger(__name__)


class QalxWorker(QalxSignalAdapter,
                 QalxUpdateStatusAdapter):
    _entity_class = Worker
    signal_class = QalxWorkerSignal

    def list_endpoint(self, *args, __bot_entity, **kwargs):
        """
        Builds the list_endpoint for workers.  This requires the bot_entity
        which will get passed down from the calling method via kwargs.  We
        mangle the name so that we can identify this specific kwarg compared to
        normal filter kwargs
        :param __bot_entity:An instance of ~entities.bot.Bot
        :type __bot_entity:~entities.bot.Bot
        :return:The Worker list endpoint
        """
        bot_endpoint = self.session.bot.detail_endpoint(guid=__bot_entity['guid'])
        return f'{bot_endpoint}/{self.entity_class.entity_type}'

    def detail_endpoint(self, guid, *args, __bot_entity, **kwargs):
        """
        Builds the list_endpoint for workers.  This requires the bot_entity
        which will get passed down from the calling method via kwargs.  We
        mangle the name so that we can identify this specific kwarg compared to
        normal filter kwargs
        :param guid:The guid of the Worker
        :type guid:guid
        :param __bot_entity:An instance of ~entities.bot.Bot
        :type __bot_entity:~entities.bot.Bot
        :return:The Worker list endpoint
        """
        # Because the list endpoint expects a mangled name we need to pass
        # the bot_entity through to a mangled parameter.
        bot_endpoint = self.list_endpoint(_QalxWorker__bot_entity=__bot_entity)
        return f'{bot_endpoint}/{guid}'

    def get(self, guid, bot_entity, *args, **kwargs):
        """
        Gets an individual worker.
        :param guid:The guid of the Worker
        :type guid: guid
        :param bot_entity:An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        """
        # The bot entity is passed through to a mangled parameter as this gets
        # sent to the `detail_endpoint` method in the base `get` method
        return super(QalxWorker, self).get(guid,
                                           _QalxWorker__bot_entity=bot_entity,
                                           *args,
                                           **kwargs)

    def find(self, bot_entity, *args, **kwargs):
        """
        Finds workers.
        :param bot_entity:An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        """
        # The bot entity is passed through to a mangled parameter as this gets
        # sent to the `list_endpoint` method in the base `get` method
        return super(QalxWorker, self).find(_QalxWorker__bot_entity=bot_entity,
                                            *args,
                                            **kwargs)

    def get_signal(self, entity, bot_entity, *args, **kwargs):
        """
        Gets the signal for a specific entity
        :param entity:An instance of ~entities.worker.Worker
        :type entity:~entities.worker.Worker
        :param bot_entity:An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        """
        # We don't pass `bot_entity` through to a mangled parameter as
        # `get_signal` calls `get` internally which handles name mangling
        return super(QalxWorker, self).get_signal(
            entity=entity,
            bot_entity=bot_entity,
            *args,
            **kwargs
        )

    def terminate(self, entity, bot_entity, *args, **kwargs):
        """
        Terminates a specific entity
        :param entity:An instance of ~entities.worker.Worker
        :type entity:~entities.worker.Worker
        :param bot_entity:An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        """
        # Don't pass `bot_entity` through to a mangled parameter as
        # `terminate` calls `get` internally which handles name mangling
        return super(QalxWorker, self).terminate(
            entity=entity,
            _QalxWorker__bot_entity=bot_entity,
            *args,
            **kwargs
        )

    def _stop_or_resume(self, entity, stop, bot_entity, **kwargs):
        """
        Stops or resumes a specific entity
        :param entity:An instance of ~entities.worker.Worker
        :type entity:~entities.worker.Worker
        :param stop:Whether we should stop or resume
        :type stop:bool
        :param bot_entity:An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        """
        # The bot entity is passed through to a mangled parameter as this gets
        # sent to the `detail_endpoint` method in the base `get` method
        return super(QalxWorker, self)._stop_or_resume(
            entity=entity,
            stop=stop,
            _QalxWorker__bot_entity=bot_entity,
            **kwargs
        )

    def reload(self, entity, bot_entity, **kwargs):
        """
        Reloads the current entity from the API
        :param entity: An instance of ~entities.worker.Worker
        :type entity:~entities.worker.Worker
        :param bot_entity: An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        :return: A refreshed instance of `self.entity`
        """
        # Don't pass `bot_entity` to a mangled parameter as the `get` method
        # handles that
        return super(QalxWorker, self).get(
            guid=entity['guid'],
            bot_entity=bot_entity
        )

    def update_status(self, entity, bot_entity, status):
        """
        Updates the workers status
        :param entity: An instance of ~entities.worker.Worker
        :type entity:~entities.worker.Worker
        :param bot_entity: An instance of ~entities.bot.Bot
        :type bot_entity:~entities.bot.Bot
        :param status: The status to update to
        :type status: str
        """
        # The bot entity is passed through to a mangled parameter as this gets
        # sent to the `detail_endpoint` method in the base `get` method
        return super(QalxWorker, self).update_status(
            entity=entity,
            _QalxWorker__bot_entity=bot_entity,
            status=status,
        )




"""pyqalx.core.errors defines QalxError exception and a load of children.

If pyqalx is going raise an error you know about then use one of these or create a new one.

"""


class QalxError(Exception):
    """Base qalx error. Take responsibility!"""
    pass


class QalxAuthError(QalxError):
    """qalx did not find a way to authenticate or the authentication didn't work"""
    pass


class QalxNoGUIDError(QalxError):
    """A QalxEntity without a guid is like a dog without a bone."""
    pass


class QalxNoInfoError(QalxError):
    """A QalxEntity without info is like a dog without a bone."""
    pass


class QalxReturnedMultipleError(QalxError):
    """There should only be one thing. But qalx sent more than one."""
    pass


class QalxConfigProfileNotFound(QalxError):
    """The profile wasn't in the file or the file wasn't properly formed."""
    pass


class QalxConfigFileNotFound(QalxError):
    """There should be a file in the users `home` directory (either a .bots or .qalx)."""
    pass


class QalxQueueError(QalxError):
    """There wasn't the correct information to connect to the remote queue. """


class QalxBotInitialisationFailed(QalxError):
    """The bot initialisation function returned something falsey."""



class QalxEntityTypeNotFound(QalxError):
    """We couldn't find the entity type you were looking for."""


class QalxEntityNotFound(QalxError):
    """We couldn't find the entity you were looking for."""


class QalxMultipleEntityReturned(QalxError):
    """We found more than one entity, but you just wanted the one hey?"""


class QalxConfigError(QalxError):
    """Something about an attempted load of config didn't work"""


class QalxAPIResponseError(QalxError):
    """There was a problem with some kind of API request."""



class QalxEntityUnchanged(QalxError):
    """Saved something which hadn't actually been changed when we thought it had."""


class QalxInvalidSession(QalxError):
    """The qalx_session argument passed to an adapter isn't a valid QalxSession instance"""


class QalxNoEntity(QalxError):
    """A user tried to access the `entity` attribute on QalxAdapter when they hadn't set an entity"""


class QalxIncorrectEntityType(QalxError):
    """The entity on the QalxAdapter is of a different type to the type of adapter"""


class QalxFileError(QalxError):
    """There is something wrong with the file details passed to the QalxAdapter"""


class QalxAlreadyRegistered(QalxError):
    """The entity is already registered with the session"""
    pass


class QalxCannotUnregister(QalxError):
    """The entity cannot be unregistered"""
    pass


class QalxRegistrationClassNotFound(QalxError):
    """The registration class for registering a custom class was not found"""
    pass


class QalxInvalidBlueprintError(QalxError):
    """The blueprint is not valid."""
    pass


class QalxInvalidTagError(QalxError):
    """The user does not have access to write to the specific tags"""
    pass


class QalxStepFunctionNotDefined(QalxError):
    """A specific Bot step function has not been defined by the user"""
    pass



import deprecation
import logging

from pyqalx.__version__ import __version__
from pyqalx.core.adapters.adapter import QalxValidateBlueprintAdapter, \
    QalxFileAdapter
from pyqalx.core.entities import Item
from pyqalx.core.entities.item import ItemAddManyEntity
from pyqalx.core.errors import QalxFileError, QalxError


logger = logging.getLogger(__name__)


class QalxItem(QalxValidateBlueprintAdapter,
               QalxFileAdapter):
    _entity_class = Item

    def _validate(self, data=None, input_file=None, file_name="",
                  meta=None, blueprint_name=None,
                  **kwargs):
        if not any([data, input_file]):
            raise QalxError('At least one of `data` or `input_file` is '
                            'required')
        if data is None:
            data = {}
        return super(QalxItem, self)._validate(data=data,
                                               input_file=input_file,
                                               file_name=file_name,
                                               meta=meta,
                                               blueprint_name=blueprint_name,
                                               **kwargs)

    def add(self, data=None, input_file=None, file_name="",
            meta=None, blueprint_name=None, **kwargs):
        """
        Adds an `Item` instance that can contain either
        data (as a dict), a file or both.

        :param data: Optional data to store against this Item
        :type data: dict
        :param input_file: Optional input file path or stream
        :type input_file: str or `filestream`
        :param file_name: input file name. Optional for file path. Required for file stream
        :type file_name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param blueprint_name: An optional blueprint name to use to validate this item against an existing Blueprint
        :return: A newly created `Item` instance
        """

        return super(QalxItem, self).add(data=data,
                                         input_file=input_file,
                                         file_name=file_name,
                                         meta=meta,
                                         blueprint_name=blueprint_name,
                                         **kwargs)

    def save(self, entity, input_file=None, file_name="", blueprint_name=None,
             **kwargs):
        """
        Saves an updated existing `Item` instance.
        To remove a file from an `Item` update the entity so that
        `entity['file'] = {}`

        :param entity: The entity that we are saving
        :type entity: An instance of `Item`
        :param input_file: Optional input file path or stream
        :type input_file: str or `filestream`
        :param file_name: input file name. Optional for file path. Required
                          for file stream
        :type file_name: str
        :param meta: A dictionary of metadata to store
        :type meta: dict
        :param blueprint_name: An optional blueprint name to use if you want
                              to validate this item against an existing
                              Blueprint
        :type blueprint_name: str
        :return: An updated `Item` instance
        """
        return super(QalxItem, self).save(entity=entity,
                                          input_file=input_file,
                                          file_name=file_name,
                                          blueprint_name=blueprint_name,
                                          **kwargs)

    def add_many(self, items, **kwargs):
        """
        Adds multiple items at once.
        :param items: A list of item data that you want to create.  This can
        be a mixture of non file items and file items. The keys of each item in the list should be the same as if a
        single item were being created via the `add` method
        :type items: list
        :return:
        """
        if not isinstance(items, list):
            raise QalxError('`items` argument must be a list')
        headers = {
            'qalx-bulk': 'true'
        }
        validated_items = []

        for item in items:
            # Ensure all the items are valid.  They should be in the same
            # format as if a user is adding a single item.  This also handles
            # checking the blueprint for each item
            validated_items.append(self._validate(**item))

        resp = self.add(items=validated_items,
                        headers=headers,
                        _is_validated=True,
                        # Don't upload as we need to handle
                        # the file upload manually
                        upload=False,
                        **kwargs)

        # We have all the data to create the files, but the indexes have
        # been split up compared to the response.  Therefore, inject blank
        # values into the response where there are errors.  This means that
        # the structure will match the original `items` that the user specified
        # (as the API preserves order when bulk creating data)
        ERRORED_KEY = None
        [resp['items'].insert(int(x), ERRORED_KEY) for x in resp['errors'].keys()]

        for index, item in enumerate(items):
            # Iterate the original `items` provided by the user
            created_item = resp['items'][index]
            if item.get('input_file') and created_item is not None:
                # There was a file on the original item and it was also
                # successfully created by the API.  Upload the file as normal
                self._upload_file(entity=resp['items'][index],
                                  input_file=item['input_file'])
        # Finally, remove all the `None` values from the items as these were
        # errors that we are done with
        resp['items'] = list(filter(lambda x: x is not ERRORED_KEY, resp['items']))
        if resp['errors']:
            for index, error in resp['errors'].items():
                logger.warning(f'Error bulk creating at index `{index}`. '
                               f'Error: `{error}`. '
                               f'Original Data: `{items[int(index)]}`')
        return ItemAddManyEntity(self.entity_class, resp)

    @deprecation.deprecated(deprecated_in="0.6", removed_in="1.0",
                            current_version=__version__,
                            details="Use the `add` function instead")
    def add_file(self, input_file, data=None, meta=None,
                 file_name="", blueprint_name=None, **kwargs):
        """adds a file to qalx with optional data and metadata

        :param input_file: input file path or stream
        :type input_file: str or `filestream`
        :param data: optional data you want to store with this file
        :type data: dict
        :param meta: additional data about the item
        :param meta: dict
        :param file_name: input file name. Optional for file path. Required
                          for file stream
        :type file_name: str
        :param blueprint_name: An optional blueprint name to use if you want
                              to validate this item against an existing
                              Blueprint
        :type blueprint_name: str
        :return: pyqalx.core.Item
        """
        if data is None:
            data = {}

        if self.session.rest_api.is_filestream(input_file) and not file_name:
            raise QalxFileError('You must supply a file name when supplying'
                                ' a file stream')
        return self.add(input_file=input_file,
                        file_name=file_name,
                        data=data,
                        meta=meta,
                        blueprint_name=blueprint_name,
                        **kwargs)



from importlib import import_module
from os import getcwd
from pprint import pprint

import click
import sys
from click import ClickException

from pyqalx import Bot, QalxSession
from pyqalx.config import UserConfig, BotConfig, QalxConfigProfileNotFound, \
    QalxConfigFileNotFound


@click.group(chain=True)
@click.option(
    '-u', '--user-profile',
    help='User profile name in .qalx.',
    default="default"
)
@click.option(
    '-b', '--bot-profile',
    help='Bot profile name in .bots.',
    default="default"
)
@click.pass_context
def qalx(ctx, user_profile, bot_profile):
    """Command line interface to qalx."""
    ctx.ensure_object(dict)
    ctx.obj['USER_PROFILE'] = user_profile
    ctx.obj['BOT_PROFILE'] = bot_profile
    sys.path.append(getcwd())


@qalx.command("start-bot")
@click.option(
    '-p', '--processes',
    help='The number of workers to spawn in the bot.',
    default=1
)
@click.option(
    '--skip-ini/--no-skip-ini',
    help='Should the bot use the profiles on disk?',
    default=False
)
@click.option(
    '--add-queue/--no-add-queue',
    help="Should the bot try and create the queue if it doesn't exist?",
    default=True
)
@click.argument("target")
@click.pass_context
def start(ctx, target, processes, skip_ini, add_queue):
    """
    start the bot at TARGET.

    This is the import path with a colon followed by the variable name of the bot. e.g. `my_qalx.my_bot:bot`
    """
    bot_module_name, bot_variable_name = target.split(":")
    bot_module = import_module(bot_module_name)

    if bot_variable_name not in dir(bot_module):
        raise ClickException(f"{bot_variable_name} not found in {bot_module_name}.")
    else:
        bot = getattr(bot_module, bot_variable_name)
        if not isinstance(bot, Bot):
            raise ClickException(f"{bot_variable_name} isn't an instance of Bot.")
        click.echo(f"Starting {bot.name} reading from {bot.queue_name} with"
                   f" {processes} workers.")
        user_profile = ctx.obj.get("USER_PROFILE", "default")
        bot_profile = ctx.obj.get("BOT_PROFILE", "default")
        bot.start(user_profile_name=user_profile,
                  bot_profile_name=bot_profile,
                  processes=int(processes),
                  skip_ini=skip_ini,
                  add_queue=add_queue)


@qalx.command("terminate-bot")
@click.argument("name")
@click.pass_context
def terminate_bot(ctx, name):
    """shutdown the bot named NAME"""
    qalx_session = QalxSession(profile_name=ctx.obj.get("USER_PROFILE", "default"))
    bot_to_kill = qalx_session.bot.find_one(name=name)
    qalx_session.bot.terminate(bot_to_kill)
    click.echo(f"Terminated {name}")


@qalx.command("stop-bot")
@click.argument("name")
@click.pass_context
def stop_bot(ctx, name):
    """stop the bot named NAME.

    All workers to stop pulling jobs from the queue.
    """
    qalx_session = QalxSession(profile_name=ctx.obj.get("USER_PROFILE", "default"))
    bot_to_stop = qalx_session.bot.find_one(name=name)
    qalx_session.bot.stop(bot_to_stop)
    click.echo(f"Stopping {name}.")


@qalx.command("resume-bot")
@click.argument("name")
@click.pass_context
def resume_bot(ctx, name):
    """resume the bot named NAME.

    All workers to start pulling jobs from the queue.
    """
    qalx_session = QalxSession(profile_name=ctx.obj.get("USER_PROFILE", "default"))
    bot_to_resume = qalx_session.bot.find_one(name=name)
    qalx_session.bot.resume(bot_to_resume)
    click.echo(f"Resuming {name}.")


@qalx.command("bot-info")
@click.argument("name")
@click.pass_context
def bot_info(ctx, name):
    """print info about the bot named NAME.

    """
    qalx_session = QalxSession(profile_name=ctx.obj.get("USER_PROFILE", "default"))
    bot_to_info = qalx_session.bot.find_one(name=name)
    click.echo(pprint(bot_to_info))


@qalx.command("terminate-workers")
@click.argument("bot_name")
@click.argument("number")
@click.pass_context
def terminate_workers(ctx, bot_name, number):
    """terminate NUMBER of workers on bot named BOT_NAME."""
    qalx_session = QalxSession(profile_name=ctx.obj.get("USER_PROFILE", "default"))
    bot_to_kill = qalx_session.bot.find_one(name=bot_name)
    if int(number) > len(bot_to_kill['workers']):
        raise ClickException(
            f"{number} is more than the number of workers on {bot_name} ({len(bot_to_kill['workers'])})")
    for n in range(int(number)):
        qalx_session.worker.terminate(bot_to_kill['workers'][n], bot_entity=bot_to_kill)
        click.echo(f"Terminated {n + 1} worker.")


@qalx.command('configure')
@click.option('--user/--no-user', default=True)
@click.option('--bot/--no-bot', default=True)
@click.argument('extra', nargs=-1)
@click.pass_context
def configure(ctx, user, bot, extra):
    """
    Configures the .qalx and .bots config files.
    Usage:

    `qalx configure` - configures user and bot default profiles

    `qalx --user-profile=dev configure` - configures dev user profile and
    default bot profile

    `qalx --bot-profile=dev configure --no-user` - only configures bot dev
    profile.  Doesn't configure user profile

    `qalx configure --no-bot customkey=customvalue customkey2=customvalue2`
    - configures user profile with two extra key/values pairs on the config
    """

    if not user and not bot:
        raise ClickException('`user` or `bot` must be specified otherwise no '
                             'config files can be created.  Either call '
                             'without any arguments or only use a '
                             'single `--no-user` or `--no-bot` switch')

    user_profile = ctx.obj["USER_PROFILE"]
    bot_profile = ctx.obj["BOT_PROFILE"]
    check_existing = []

    if user:
        check_existing.append((UserConfig, user_profile, 'user'),)
    if bot:
        check_existing.append((BotConfig, bot_profile, 'bot'))

    for ConfigClass, profile_name, config_type in check_existing:
        # If the profiles already exist we just exit.  It's up to the user
        # to make any changes manually to the profile file - otherwise we risk
        # parsing the file incorrectly and overwriting data the user wants to
        # keep
        try:
            ConfigClass().from_inifile(profile_name)
            config_path = ConfigClass.config_path()
            raise ClickException(f'`{profile_name}` profile for config stored '
                                 f'at `{config_path}` '
                                 f'already exists.  To make changes you must '
                                 f'edit the config file directly or specify a '
                                 f'`--no-{config_type}` flag')
        except (QalxConfigProfileNotFound, QalxConfigFileNotFound):
            # QalxConfigProfileNotFound: The specific profile could not be
            #                            found in the given ConfigClass file
            # QalxConfigFileNotFound: The specific file could not be found
            #                         for the given ConfigClass
            pass

    config_items = {}

    for item in extra:
        # Handles the user being able to pass in extra key/value pairs to write
        # to the config file
        split_item = item.split('=', 1)
        if len(split_item) != 2:
            raise ClickException(f'extra arguments must be in the format '
                                 f'`key=value`.  Got `{item}`')
        config_items[split_item[0].upper()] = split_item[1]

    if user:
        value = click.prompt('qalx Token', type=str)
        config_items['TOKEN'] = value
        UserConfig.configure(user_profile, config_items)
    if bot:
        BotConfig.configure(bot_profile)



from .blueprint import QalxBlueprint
from .bot import QalxBot
from .group import QalxGroup
from .item import QalxItem
from .queue import QalxQueue
from .set import QalxSet
from .worker import QalxWorker



"""
Core API
~~~~~~~~

.. automodule:: pyqalx.core
    :members:

Bot API
~~~~~~~

.. automodule:: pyqalx.bot
    :members:

"""
from .core.adapters import QalxItem, \
    QalxSet, QalxGroup, QalxQueue, QalxBot
from .core.session import QalxSession
from .bot import Bot
from .core.entities import Blueprint, Item, Set, Group, Queue

